# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass, field

from dataclasses_json.undefined import CatchAll

from faebryk.libs.kicad.fileformats_common import C_stroke, C_xy, C_xyr
from faebryk.libs.kicad.fileformats_latest import (
    C_arc,
    C_footprint,
    C_fp_text,
    C_kicad_footprint_file,
)
from faebryk.libs.kicad.fileformats_sch import C_circle as C_symbol_Circle
from faebryk.libs.kicad.fileformats_sch import C_symbol
from faebryk.libs.kicad.fileformats_v5 import (
    C_circle_v5,
    C_line_v5,
    C_rect_v5,
)
from faebryk.libs.sexp.dataclass_sexp import SEXP_File, sexp_field

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class C_arc_v6:
    start: C_xy
    mid: C_xy
    end: C_xy
    # instead of stroke
    width: float
    layer: str
    # no uuid

    def convert_to_new(self) -> C_arc:
        return C_arc(
            start=self.start,
            mid=self.mid,
            end=self.end,
            stroke=C_stroke(width=self.width, type=C_stroke.E_type.default),
            layer=self.layer,
        )


@dataclass(kw_only=True)
class C_fp_text_v6(C_fp_text):
    @dataclass(kw_only=True)
    class C_xyUnlocked:
        x: float = field(**sexp_field(positional=True))
        y: float = field(**sexp_field(positional=True))
        unknown: CatchAll = None

    at: C_xyUnlocked

    def convert_to_new(self) -> C_fp_text:
        return C_fp_text(
            type=self.type,
            text=self.text,
            at=C_xyr(x=self.at.x, y=self.at.y),
            layer=self.layer,
            uuid=self.uuid,
            effects=self.effects,
            # TODO its inside the at
            unlocked=False,
        )


@dataclass
class C_kicad_footprint_file_v6(SEXP_File):
    @dataclass(kw_only=True)
    class C_footprint_in_file(C_kicad_footprint_file.C_footprint_in_file):
        fp_lines: list[C_line_v5] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        fp_arcs: list[C_arc_v6] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        fp_circles: list[C_circle_v5] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        fp_rects: list[C_rect_v5] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        fp_texts: list[C_fp_text_v6] = field(
            **sexp_field(multidict=True), default_factory=list
        )

        version: int = 20211014

        def convert_to_new(self) -> C_kicad_footprint_file.C_footprint_in_file:
            propertys: dict[str, C_footprint.C_property] = {
                name: C_footprint.C_property(
                    name=name,
                    value=k.text,
                    at=C_xyr(x=k.at.x, y=k.at.y),
                    layer=k.layer,
                    uuid=k.uuid,
                    effects=k.effects,
                )
                for k in self.fp_texts
                if (name := k.type.capitalize()) in ("Reference", "Value")
            } | self.propertys

            texts = [t for t in self.fp_texts if t.type not in ("reference", "value")]
            for t in self.fp_texts:
                if t.type == "reference":
                    texts.append(
                        C_fp_text_v6(
                            type=C_fp_text.E_type.user,
                            text=t.text.replace("REF**", "${REFERENCE}"),
                            at=t.at,
                            layer=t.layer,
                            uuid=t.uuid,
                            effects=t.effects,
                        )
                    )

            return C_kicad_footprint_file.C_footprint_in_file(
                fp_lines=[line.convert_to_new() for line in self.fp_lines],
                fp_arcs=[arc.convert_to_new() for arc in self.fp_arcs],
                fp_circles=[circle.convert_to_new() for circle in self.fp_circles],
                fp_rects=[rect.convert_to_new() for rect in self.fp_rects],
                # fp-file
                tedit=self.tedit,
                descr=self.descr or "",
                tags=self.tags,
                generator="faebryk",
                generator_version="v6",
                # fp
                name=self.name,
                layer=self.layer,
                propertys=propertys,
                attr=self.attr,
                fp_texts=[t.convert_to_new() for t in texts],
                fp_poly=self.fp_poly,
                pads=self.pads,
                models=self.models,
            )

    footprint: C_footprint_in_file

    def convert_to_new(self) -> C_kicad_footprint_file:
        return C_kicad_footprint_file(footprint=self.footprint.convert_to_new())


@dataclass(kw_only=True)
class C_symbol_Circle_v6(C_symbol_Circle):
    radius: float
    end: C_xy = field(default_factory=lambda: C_xy(0, 0))

    def convert_to_new(self) -> C_symbol_Circle:
        return C_symbol_Circle(
            center=self.center,
            # TODO does this work?
            end=C_xy(self.center.x + self.radius, self.center.y),
            stroke=self.stroke,
            fill=self.fill,
        )


@dataclass(kw_only=True)
class C_symbol_symbol_v6(C_symbol.C_symbol):
    circles: list[C_symbol_Circle_v6] = field(
        **sexp_field(multidict=True), default_factory=list
    )

    def convert_to_new(self) -> C_symbol.C_symbol:
        return C_symbol.C_symbol(
            name=self.name,
            polylines=self.polylines,
            circles=[circle.convert_to_new() for circle in self.circles],
            rectangles=self.rectangles,
            arcs=self.arcs,
            pins=self.pins,
        )


@dataclass(kw_only=True)
class C_symbol_v6(C_symbol):
    symbols: dict[str, C_symbol_symbol_v6] = field(
        **sexp_field(multidict=True, key=lambda x: x.name),
        default_factory=dict,
    )

    def convert_to_new(self) -> C_symbol:
        return C_symbol(
            name=self.name,
            power=self.power,
            propertys=self.propertys,
            pin_numbers=self.pin_numbers,
            pin_names=self.pin_names,
            in_bom=self.in_bom,
            on_board=self.on_board,
            symbols={
                name: symbol.convert_to_new() for name, symbol in self.symbols.items()
            },
            convert=self.convert,
        )


@dataclass(kw_only=True)
class C_symbol_in_file_v6(SEXP_File):
    symbol: C_symbol_v6



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import textwrap

import pytest

from faebryk.libs.util import md_list


def _fmt(s: str) -> str:
    return textwrap.dedent(s).strip()


@pytest.mark.parametrize(
    "obj,recursive,expected",
    [
        # Empty collections
        ([], False, "- *(empty)*"),
        ({}, False, "- *(empty)*"),
        # Simple list
        (
            [1, 2, 3],
            False,
            _fmt(
                """
                - 1
                - 2
                - 3
                """
            ),
        ),
        # Simple list with recursive flag
        (
            [1, 2, 3],
            True,
            _fmt(
                """
                - 1
                - 2
                - 3
                """
            ),
        ),
        # Simple dictionary
        (
            {"a": 1, "b": 2},
            False,
            _fmt(
                """
                - **a:** 1
                - **b:** 2
                """
            ),
        ),
        # Nested list, non-recursive
        (
            [1, [2, 3]],
            False,
            _fmt(
                """
                - 1
                - [2, 3]
                """
            ),
        ),
        # Nested list, recursive
        (
            [1, [2, 3]],
            True,
            _fmt(
                """
                - 1
                  - 2
                  - 3
                """
            ),
        ),
        # Nested dictionary, non-recursive
        (
            {"a": 1, "b": {"c": 2}},
            False,
            _fmt(
                """
                - **a:** 1
                - **b:** {'c': 2}
                """
            ),
        ),
        # Nested dictionary, recursive
        (
            {"a": 1, "b": {"c": 2}},
            True,
            _fmt(
                """
                - **a:** 1
                - **b:**
                  - **c:** 2
                """
            ),
        ),
        # Mixed nested structures, recursive
        (
            {"a": 1, "b": [2, 3]},
            True,
            _fmt(
                """
                - **a:** 1
                - **b:**
                  - 2
                  - 3
                """
            ),
        ),
        # Non-iterable object
        (123, False, "- 123"),
        # String
        (
            "abc",
            False,
            "- abc",
        ),
        # Deep nesting, recursive
        (
            [1, [2, [3, 4]]],
            True,
            _fmt(
                """
                - 1
                  - 2
                    - 3
                    - 4
                """
            ),
        ),
        # List of dicts of lists
        # TODO: currently not supported, but also not needed yet
        # (
        #     [{"a": [1, 2, 3], "b": [4, 5, 6]}, {"c": [7, 8, 9]}],
        #     False,
        #     _fmt(
        #         """
        #         - **a:**
        #           - 1
        #           - 2
        #           - 3
        #         - **b:**
        #           - 4
        #           - 5
        #           - 6
        #         - **c:**
        #           - 7
        #           - 8
        #           - 9
        #         """
        #     ),
        # ),
    ],
)
def test_md_list(obj, recursive, expected):
    assert md_list(obj, recursive=recursive) == expected



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import json
import logging
from enum import StrEnum

import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.module import Module
from faebryk.core.parameter import Parameter
from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer
from faebryk.libs.picker.lcsc import PickedPartLCSC
from faebryk.libs.picker.lcsc import attach as lcsc_attach
from faebryk.libs.sets.sets import P_Set
from faebryk.libs.util import KeyErrorNotFound

NO_LCSC_DISPLAY = "No LCSC number"

logger = logging.getLogger(__name__)


class Properties(StrEnum):
    manufacturer = "Manufacturer"
    partno = "Partnumber"
    lcsc = "LCSC"
    param_prefix = "PARAM_"
    # used in transformer
    param_wildcard = "PARAM_*"


def load_part_info_from_pcb(G: Graph):
    """
    Load descriptive properties from footprints and saved parameters.
    """
    nodes = GraphFunctions(G).nodes_with_trait(
        PCB_Transformer.has_linked_kicad_footprint
    )

    for node, trait in nodes:
        assert isinstance(node, Module)
        if isinstance(node, F.Footprint):
            continue
        if node.has_trait(F.has_part_picked):
            continue
        assert node.has_trait(F.has_descriptive_properties), "Should load when linking"

        part_props = [Properties.lcsc, Properties.manufacturer, Properties.partno]
        fp = trait.get_fp()
        fp_props = {k.value: v for k in part_props if (v := fp.try_get_property(k))}
        if fp_props.get(Properties.lcsc) == NO_LCSC_DISPLAY:
            del fp_props[Properties.lcsc]
        props = node.get_trait(F.has_descriptive_properties).get_properties()

        # check if node has changed
        if any(props.get(k.value) != fp_props.get(k.value) for k in part_props):
            continue

        lcsc_id = props.get(Properties.lcsc)
        manufacturer = props.get(Properties.manufacturer)
        partno = props.get(Properties.partno)

        # Load Part from PCB
        if lcsc_id and manufacturer and partno:
            node.add(
                F.has_part_picked(
                    PickedPartLCSC(
                        supplier_partno=lcsc_id,
                        manufacturer=manufacturer,
                        partno=partno,
                    )
                )
            )
        elif lcsc_id:
            node.add(
                F.has_explicit_part.by_supplier(
                    supplier_partno=lcsc_id,
                    supplier_id="lcsc",
                )
            )
        elif manufacturer and partno:
            node.add(
                F.has_explicit_part.by_mfr(
                    mfr=manufacturer,
                    partno=partno,
                )
            )

        if lcsc_id:
            lcsc_attach(node, lcsc_id)

        if "Datasheet" in fp_props:
            node.add(F.has_datasheet_defined(fp_props["Datasheet"]))

        # Load saved parameters from descriptive properties
        for key, value in props.items():
            if not key.startswith(Properties.param_prefix):
                continue

            param_name = key.removeprefix(Properties.param_prefix)
            # Skip if parameter doesn't exist in node
            try:
                param = node[param_name]
            except KeyErrorNotFound:
                logger.warning(
                    f"Parameter {param_name} not found in node {node.get_name()}"
                )
                continue
            param_value = json.loads(value)
            param_value = P_Set.deserialize(param_value)
            assert isinstance(param, Parameter)
            param.alias_is(param_value)


def save_part_info_to_pcb(G: Graph):
    """
    Save parameters to footprints (by copying them to descriptive properties).
    """

    nodes = GraphFunctions(G).nodes_with_trait(F.has_part_picked)

    for node, _ in nodes:
        if t := node.try_get_trait(F.has_part_picked):
            part = t.get_part()
            if isinstance(part, PickedPartLCSC):
                node.add(
                    F.has_descriptive_properties_defined(
                        {Properties.lcsc: part.lcsc_id}
                    )
                )
                # TODO save info?
            node.add(
                F.has_descriptive_properties_defined(
                    {
                        Properties.manufacturer: part.manufacturer,
                        Properties.partno: part.partno,
                    }
                )
            )

        for p in node.get_children(direct_only=True, types=Parameter):
            lit = p.try_get_literal()
            if lit is None:
                continue
            lit = P_Set.from_value(lit)
            key = f"{Properties.param_prefix}{p.get_name()}"
            value = json.dumps(lit.serialize())
            node.add(F.has_descriptive_properties_defined({key: value}))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.units import P, Quantity
from faebryk.libs.util import times


class SOIC(F.Footprint):
    def __init__(
        self,
        pin_cnt: int,
        size_xy: tuple[Quantity, Quantity],
        pitch: Quantity,
    ) -> None:
        super().__init__()
        self._pin_cnt = pin_cnt
        self._size_xy = size_xy
        self._pitch = pitch

    @L.rt_field
    def pins(self):
        return times(self._pin_cnt, F.Pad)

    class _has_kicad_footprint(F.has_kicad_footprint_equal_ifs):
        def get_kicad_footprint(self) -> str:
            obj = self.obj
            assert isinstance(obj, SOIC)
            return "Package_SO:SOIC-{leads}_{size_x:.1f}x{size_y:.1f}mm_P{pitch:.2f}mm".format(  # noqa: E501
                leads=obj._pin_cnt,
                size_x=obj._size_xy[0].to(P.mm).m,
                size_y=obj._size_xy[1].to(P.mm).m,
                pitch=obj._pitch.to(P.mm).m,
            )

    kicad_footprint: _has_kicad_footprint
    attach_via_pinmap: F.can_attach_via_pinmap_equal
    equal_pins: F.has_equal_pins_in_ifs



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P
from faebryk.libs.util import assert_once  # noqa: F401

logger = logging.getLogger(__name__)


class Analog_Devices_ADM2587EBRWZ(Module):
    """
    Signal and power isolated RS-485 full/half-duplex transceiver with
    ±15 kV ESD protection
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    power_unisolated: F.ElectricPower
    power_isolated_out: F.ElectricPower
    power_isolated_in: F.ElectricPower
    uart: F.UART_Base
    read_enable: F.ElectricLogic
    write_enable: F.ElectricLogic
    rs485: F.RS485HalfDuplex

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    explicit_part = L.f_field(F.has_explicit_part.by_supplier)("C12081")
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def can_attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            pinmap={
                "1": self.power_unisolated.lv,
                "2": self.power_unisolated.hv,
                "3": self.power_unisolated.lv,
                "4": self.uart.rx.line,
                "5": self.read_enable.line,
                "6": self.write_enable.line,
                "7": self.uart.tx.line,
                "8": self.power_unisolated.hv,
                "9": self.power_unisolated.lv,
                "10": self.power_unisolated.lv,
                "11": self.power_isolated_out.lv,
                "12": self.power_isolated_out.hv,
                "13": self.rs485.diff_pair.p.line,
                "14": self.power_isolated_out.lv,
                "15": self.rs485.diff_pair.n.line,
                "16": self.power_isolated_out.lv,
                "17": self.rs485.diff_pair.n.line,
                "18": self.rs485.diff_pair.p.line,
                "19": self.power_isolated_in.hv,
                "20": self.power_isolated_out.lv,
            }
        )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.rs485.diff_pair.p.line: ["A"],
                self.rs485.diff_pair.n.line: ["B"],
                self.write_enable.line: ["DE"],
                self.power_unisolated.lv: ["GND1"],
                self.power_isolated_out.lv: ["GND2"],
                self.read_enable.line: ["RE#"],
                self.uart.rx.line: ["RXD"],
                self.uart.tx.line: ["TXD"],
                self.power_unisolated.hv: ["VCC"],
                self.power_isolated_in.hv: ["VISOIN"],
                self.power_isolated_out.hv: ["VISOOUT"],
                self.rs485.diff_pair.p.line: ["Y"],
                self.rs485.diff_pair.n.line: ["Z"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __init__(self, full_duplex: bool = False):
        super().__init__()
        self._full_duplex = full_duplex
        if full_duplex:
            raise NotImplementedError("Full duplex RS485 not implemented")

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.power_isolated_out.voltage.constrain_subset(
            L.Range.from_center_rel(3.3 * P.V, 0.1)
        )
        self.power_unisolated.voltage.constrain_subset(L.Range(3.3 * P.V, 5 * P.V))

        F.ElectricLogic.connect_all_module_references(
            self,
            exclude=[
                self.power_unisolated,
                self.uart,
                self.read_enable,
                self.write_enable,
            ],
        )

        # TODO: ugly
        self.rs485.diff_pair.n.reference.connect(self.power_isolated_out)
        self.rs485.diff_pair.p.reference.connect(self.power_isolated_out)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.smd import SMDSize
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class B0505S_1WR3_ReferenceDesign(Module):
    ic: F.B0505S_1WR3

    power_in: F.ElectricPower
    power_out: F.ElectricPower

    ferrite_in: F.Inductor
    ferrite_out: F.Inductor

    anti_emi_capacitor: F.Capacitor

    def __preinit__(self):
        self.power_in.voltage.alias_is(self.ic.power_in.voltage)
        self.power_out.voltage.alias_is(self.ic.power_out.voltage)

        self.ferrite_in.inductance.constrain_subset(
            L.Range.from_center_rel(4.7 * P.uH, 0.1)
        )
        self.ferrite_out.inductance.constrain_subset(
            L.Range.from_center_rel(4.7 * P.uH, 0.1)
        )
        # TODO: remove when picker is implemented
        self.ferrite_in.add(F.has_explicit_part.by_supplier("C394952"))
        self.ferrite_out.add(F.has_explicit_part.by_supplier("C394952"))

        self.power_in.decoupled.decouple(owner=self).explicit(
            nominal_capacitance=4.7 * P.uF,
            tolerance=0.2,
            size=SMDSize.I0805,
        )
        self.ic.power_in.decoupled.decouple(owner=self).explicit(
            nominal_capacitance=4.7 * P.uF,
            tolerance=0.2,
            size=SMDSize.I0805,
        )
        self.power_out.decoupled.decouple(owner=self).explicit(
            nominal_capacitance=10 * P.uF,
            tolerance=0.2,
            size=SMDSize.I0805,
        )
        self.anti_emi_capacitor.capacitance.constrain_subset(
            L.Range.from_center_rel(270 * P.pF, 0.2)
        )
        self.anti_emi_capacitor.max_voltage.constrain_subset(L.Range(min=2 * P.kV))
        # TODO: Y5P not yet supported by picker
        self.anti_emi_capacitor.add(F.has_explicit_part.by_supplier("C2914698"))

        self.power_in.lv.connect_via(self.anti_emi_capacitor, self.power_out.lv)

        self.power_in.hv.connect_via(self.ferrite_in, self.ic.power_in.hv)
        self.power_in.lv.connect(self.ic.power_in.lv)

        self.power_out.hv.connect_via(self.ferrite_out, self.ic.power_out.hv)
        self.power_out.lv.connect(self.ic.power_out.lv)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


class ufds:
    parent_node = {}
    rank = {}

    def make_set(self, u):
        for i in u:
            self.parent_node[i] = i
            self.rank[i] = 0

    def op_find(self, k):
        if self.parent_node[k] != k:
            self.parent_node[k] = self.op_find(self.parent_node[k])
        return self.parent_node[k]

    def op_union(self, a, b):
        x = self.op_find(a)
        y = self.op_find(b)

        if x == y:
            return
        if self.rank[x] > self.rank[y]:
            self.parent_node[y] = x
        elif self.rank[x] < self.rank[y]:
            self.parent_node[x] = y
        else:
            self.parent_node[x] = y
            self.rank[y] = self.rank[y] + 1



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


from typing import Mapping

import faebryk.library._F as F
from faebryk.core.node import Node
from faebryk.core.trait import TraitImpl


class has_descriptive_properties_defined(F.has_descriptive_properties.impl()):
    def __init__(self, properties: Mapping[str, str]) -> None:
        super().__init__()
        self.properties: dict[str, str] = dict(properties)

    def get_properties(self) -> dict[str, str]:
        return self.properties

    def handle_duplicate(self, old: TraitImpl, node: Node) -> bool:
        if not isinstance(old, has_descriptive_properties_defined):
            assert isinstance(old, F.has_descriptive_properties)
            self.properties.update(old.get_properties())
            return super().handle_duplicate(old, node)

        old.properties.update(self.properties)
        return False



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import logging
from enum import Enum, auto
from typing import Any, Callable, Sequence

from more_itertools import first

import faebryk.library._F as F
from faebryk.core.graph import GraphFunctions
from faebryk.core.node import Node
from faebryk.core.trait import Trait

logger = logging.getLogger(__name__)


class implements_design_check(Trait.TraitT.decless()):
    class CheckStage(Enum):
        POST_DESIGN = auto()
        POST_SOLVE = auto()
        POST_PCB = auto()

    class UnfulfilledCheckException(Exception):
        nodes: Sequence[Node]

        def __init__(self, message: str, nodes: Sequence[Node]):
            self.nodes = nodes
            super().__init__(message)

    class MaybeUnfulfilledCheckException(UnfulfilledCheckException): ...

    # TODO: make this decorators also create the implements_design_check trait
    # consider using an rt_field for that

    # Simple decorators, that are for now only used for type checking
    @staticmethod
    def register_post_design_check(func: Callable[[Any], None]):
        if not func.__name__ == "__check_post_design__":
            raise TypeError(f"Method {func.__name__} is not a post-design check name.")
        return func

    @staticmethod
    def register_post_solve_check(func: Callable[[Any], None]):
        """
        Guarantees solver availability via F.has_solver.find_unique
        """
        if not func.__name__ == "__check_post_solve__":
            raise TypeError(f"Method {func.__name__} is not a post-solve check name.")
        return func

    @staticmethod
    def register_post_pcb_check(func: Callable[[Any], None]):
        """
        Guarantees PCB availability via Node in Graph
        """
        if not func.__name__ == "__check_post_pcb__":
            raise TypeError(f"Method {func.__name__} is not a post-pcb check name.")
        return func

    def get_solver(self):
        return F.has_solver.find_unique(self.get_graph()).solver

    def get_pcb(self):
        from faebryk.library.PCB import PCB

        matches = GraphFunctions(self.get_graph()).nodes_of_type(PCB)
        assert len(matches) == 1
        return first(matches)

    def check_post_design(self):
        if not hasattr(self.get_obj(Trait), "__check_post_design__"):
            return False
        self.get_obj(Trait).__check_post_design__()  # type: ignore
        return True

    def check_post_solve(self):
        if not hasattr(self.get_obj(Trait), "__check_post_solve__"):
            return False
        self.get_obj(Trait).__check_post_solve__()  # type: ignore
        return True

    def check_post_pcb(self):
        if not hasattr(self.get_obj(Trait), "__check_post_pcb__"):
            return False
        self.get_obj(Trait).__check_post_pcb__()  # type: ignore
        return True

    def run(self, stage: CheckStage) -> bool:
        label = (
            f"`{self.get_name_of_test()}` for `{self.get_obj(Trait).get_full_name()}`"
        )
        match stage:
            case implements_design_check.CheckStage.POST_DESIGN:
                logger.debug(f"Running post-design check {label}")
                return self.check_post_design()
            case implements_design_check.CheckStage.POST_SOLVE:
                logger.debug(f"Running post-solve check {label}")
                return self.check_post_solve()
            case implements_design_check.CheckStage.POST_PCB:
                logger.debug(f"Running post-pcb check {label}")
                return self.check_post_pcb()

    def on_check(self):
        obj = self.get_obj(Trait)
        if (
            not hasattr(obj, "__check_post_design__")
            and not hasattr(obj, "__check_post_solve__")
            and not hasattr(obj, "__check_post_pcb__")
        ):
            raise TypeError(f"Trait implementation {obj} has no check methods.")

    def get_name_of_test(self) -> str:
        obj = self.get_obj(Trait)
        return type(obj).__qualname__



import pytest

from faebryk.libs.geometry.basic import Geometry


@pytest.mark.parametrize(
    "structure, axis, angle, expected",
    [
        # around origin
        ((1, 0), (0, 0), 90, (0, 1)),
        ((1, 2), (0, 0), 90, (-2, 1)),
        ((1, 2), (0, 0), 180, (-1, -2)),
        # around point
        ((0, 0), (1, 0), 180, (2, 0)),
        ((4, 0), (3, 0), 180, (2, 0)),
    ],
)
def test_rotate(structure, axis, angle, expected):
    rotated = Geometry.rotate(axis, [structure], angle)
    # rotated = tuple(map(float, rotated[0]))
    assert rotated[0] == pytest.approx(expected, abs=1e-6)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class XL_3528RGBW_WS2812B(Module):
    class _ws2812b_esphome_config(F.has_esphome_config.impl()):
        update_interval = L.p_field(units=P.s, tolerance_guess=0)

        def get_config(self) -> dict:
            obj = self.get_obj(XL_3528RGBW_WS2812B)

            data_pin = F.is_esphome_bus.find_connected_bus(obj.di.line)

            return {
                "light": [
                    {
                        "platform": "esp32_rmt_led_strip",
                        "update_interval": self.update_interval,
                        "num_leds": 1,  # TODO: make dynamic
                        "rmt_channel": 0,  # TODO: make dynamic
                        "chipset": "WS2812",
                        "rgb_order": "RGB",
                        "is_rgbw": "true",
                        "pin": data_pin.get_trait(F.is_esphome_bus).get_bus_id(),
                    }
                ]
            }

    # interfaces

    power: F.ElectricPower
    do: F.ElectricLogic
    di: F.ElectricLogic

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.LED
    )

    # Add bridge trait
    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.di, self.do)

    @L.rt_field
    def attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.power.lv,
                "2": self.di.line,
                "3": self.power.hv,
                "4": self.do.line,
            }
        )

    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://wmsc.lcsc.com/wmsc/upload/file/pdf/v2/lcsc/2402181504_XINGLIGHT-XL-3528RGBW-WS2812B_C2890364.pdf"
    )

    esphome_config: _ws2812b_esphome_config

    def __preinit__(self):
        # ------------------------------
        #          parameters
        # ------------------------------

        # ------------------------------
        #          connections
        # ------------------------------
        # FIXME
        # self.power.decoupled.decouple()
        F.ElectricLogic.connect_all_module_references(self, gnd_only=True)
        F.ElectricLogic.connect_all_module_references(self, exclude=[self.power])



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Generator, Iterable, Mapping

import faebryk.library._F as F
from atopile.errors import UserException
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import NodeNoParent
from faebryk.exporters.netlist.netlist import FBRKNetlist
from faebryk.libs.library import L
from faebryk.libs.util import FuncDict, KeyErrorAmbiguous, groupby

logger = logging.getLogger(__name__)


class can_represent_kicad_footprint(F.Footprint.TraitT.decless()):
    kicad_footprint = FBRKNetlist.Component

    def __init__(self, component: Module, graph: Graph) -> None:
        """
        graph has to be electrically closed
        """

        super().__init__()
        self.component = component
        self.graph = graph

    def get_name_and_value(self):
        return ensure_ref_and_value(self.component)

    def get_pin_name(self, pin: F.Pad):
        return self.obj.get_trait(F.has_kicad_footprint).get_pin_names()[pin]

    def get_kicad_obj(self):
        fp = self.get_obj(F.Footprint)

        properties = {
            "footprint": fp.get_trait(F.has_kicad_footprint).get_kicad_footprint()
        }

        # TODO not sure this is needed, also doing similar stuff elsewhere
        for c in [fp, self.component]:
            if c.has_trait(F.has_descriptive_properties):
                properties.update(
                    c.get_trait(F.has_descriptive_properties).get_properties()
                )

        properties["atopile_address"] = self.component.get_full_name()

        name, value = self.get_name_and_value()

        return can_represent_kicad_footprint.kicad_footprint(
            name=name,
            properties=properties,
            value=value,
        )


def ensure_ref_and_value(c: Module):
    value = (
        c.get_trait(F.has_simple_value_representation).get_value()
        if c.has_trait(F.has_simple_value_representation)
        else type(c).__name__
    )

    # At this point, all components MUST have a designator
    return c.get_trait(F.has_designator).get_designator(), value


def add_or_get_nets(*interfaces: F.Electrical):
    buses = ModuleInterface._group_into_buses(interfaces)
    nets_out = set()

    for bus_repr, connected_mifs in buses.items():
        nets_on_bus = {
            net
            for mif in connected_mifs
            if (net := F.Net.find_from_part_of_mif(mif)) is not None
        }

        if not nets_on_bus:
            net = F.Net()
            net.part_of.connect(bus_repr)
            nets_on_bus = {net}

        if len(nets_on_bus) > 1:
            raise KeyErrorAmbiguous(list(nets_on_bus), "Multiple nets interconnected")

        nets_out |= nets_on_bus

    return nets_out


def attach_nets(G: Graph) -> set[F.Net]:
    """Create nets for all the pads in the graph."""
    pad_mifs = [pad.net for pad in GraphFunctions(G).nodes_of_type(F.Pad)]
    nets = add_or_get_nets(*pad_mifs)
    return nets


# FIXME: this belongs at most in the KiCAD netlist generator
# and should likely just return the properties rather than mutating the graph
@dataclass
class _NetName:
    base_name: str | None = None
    prefix: str | None = None
    suffix: int | None = None

    @property
    def name(self) -> str:
        """
        Get the name of the net.
        Net names should take the form of: <prefix>-<base_name>-<suffix>
        There must always be some base, and if it's not provided, it's just 'net'
        Prefixes and suffixes are joined with a "-" if they exist.
        """
        return "-".join(
            str(n) for n in [self.prefix, self.base_name or "net", self.suffix] if n
        )


def _conflicts(
    names: Mapping[F.Net, _NetName],
) -> Generator[Iterable[F.Net], None, None]:
    for items in groupby(names.items(), lambda it: it[1].name).values():
        if len(items) > 1:
            yield [net for net, _ in items]


def _name_shittiness(name: str | None) -> float:
    """Caesar says 👎"""

    # These are completely shit names that
    # have no business existing
    if name is None:
        return 0

    if name == "net":
        # By the time we're here, we have a bunch
        # of pads with the name net attached
        return 0

    if name == "":
        return 0

    if name.startswith("unnamed"):
        return 0

    if re.match(r"^(_\d+|p\d+)$", name):
        return 0

    # Here are some shitty patterns, that are
    # fine as a backup, but should be avoided

    # Anything that starts with an underscore
    # is probably a temporary name
    if name.startswith("_"):
        return 0.2

    # "hv" is common from power interfaces, but
    # if there's something else available, prefer that
    if name == "hv":
        return 0.3

    # "line" is common from signal interfaces, but
    # if there's something else available, prefer that
    if name == "line":
        return 0.3

    # Anything with a trailing number is
    # generally less interesting
    if re.match(r".*\d+$", name):
        return 0.5

    return 1


def attach_net_names(nets: Iterable[F.Net]) -> None:
    """
    Generate good net names, assuming that we're passed all the nets in a design
    """

    names = FuncDict[F.Net, _NetName]()

    # Ignore nets with names already
    unnamed_nets = [n for n in nets if not n.has_trait(F.has_overriden_name)]

    # Capture already-named nets for conflict checking
    for net in nets:
        if net.has_trait(F.has_overriden_name):
            names[net] = _NetName(
                base_name=net.get_trait(F.has_overriden_name).get_name()
            )

    # First generate candidate base names
    def _decay(depth: int) -> float:
        return 1 / (depth + 1)

    # FIXME: overriden names will be modified if multiple nets are in conflict
    # FIXME: the errors for this deserve vast improvement. Attaching an origin trait
    # to has_net_name is a start, but we need a generic way to raise those as
    # well-formed errors
    for net in unnamed_nets:
        net_required_names: set[str] = set()
        net_suggested_names: list[tuple[str, int]] = []
        implicit_name_candidates: Mapping[str, float] = defaultdict(float)
        case_insensitive_map: Mapping[str, str] = {}

        for mif in net.get_connected_interfaces():
            # If there's net info, use it
            depth = len(mif.get_hierarchy())

            if t := mif.try_get_trait(F.has_net_name):
                if t.level == F.has_net_name.Level.EXPECTED:
                    net_required_names.add(t.name)
                elif t.level == F.has_net_name.Level.SUGGESTED:
                    net_suggested_names.append((t.name, len(mif.get_hierarchy())))
                continue

            # Rate implicit names
            # lower case so we are not case sensitive
            try:
                name = mif.get_name()
            except NodeNoParent:
                # Skip no names
                continue

            lower_name = name.lower()
            case_insensitive_map[lower_name] = name

            if mif.get_parent_of_type(L.ModuleInterface):
                # Give interfaces on the same level a fighting chance
                depth -= 1

            implicit_name_candidates[case_insensitive_map[lower_name]] += _decay(
                depth
            ) * _name_shittiness(lower_name)

        # Check required names
        if net_required_names:
            if len(set(net_required_names)) > 1:
                raise UserException(
                    f"Multiple conflicting required net names: {net_required_names}"
                )
            net.add(F.has_overriden_name_defined(net_required_names.pop()))
            continue

        # Initialize the net name for the remaining processing
        names[net] = _NetName()

        if net_suggested_names:
            names[net].base_name = min(net_suggested_names, key=lambda x: x[1])[0]

        elif implicit_name_candidates:
            # Type ignored on this because they typing on both max and defaultdict.get
            # is poor. This is actually correct, and supposed to return None sometimes
            names[net].base_name = max(
                implicit_name_candidates,
                key=implicit_name_candidates.get,  # type: ignore
            )

    # Resolve as many conflict as possible by prefixing on
    # the lowest common node's full name
    for conflict_nets in _conflicts(names):
        for net in conflict_nets:
            if lcn := L.Node.nearest_common_ancestor(*net.get_connected_interfaces()):
                names[net].prefix = lcn[0].get_full_name()

    # Resolve remaining conflicts by suffixing on a number
    for conflict_nets in _conflicts(names):
        for i, net in enumerate(conflict_nets):
            names[net].suffix = i

    # Override the net names we've derived
    for net, name in names.items():
        # Limit name length to 255 chars
        if len(name.name) > 255:
            name_str = name.name[:200] + "..." + name.name[-50:]
        else:
            name_str = name.name
        net.add(F.has_overriden_name_defined(name_str))

    assert all(n.has_trait(F.has_overriden_name) for n in nets)
